(* $I1: Unison file synchronizer: src/fileutil.mli $ *)
(* $I2: Last modified by zheyang on Wed, 12 Dec 2001 02:26:21 -0500 $ *)
(* $I3: Copyright 1999-2002 (see COPYING for details) $ *)

val bs2fs : string -> string
val removeTrailingSlashes : string -> string
