(* $I1: Unison file synchronizer: src/uigtk.mli $ *)
(* $I2: Last modified by vouillon on Tue, 30 May 2000 18:27:30 -0400 $ *)
(* $I3: Copyright 1999-2002 (see COPYING for details) $ *)

module Body : Uicommon.UI
