(* $I1: Unison file synchronizer: src/strings.mli $ *)
(* $I2: Last modified by bcpierce on Wed, 18 Oct 2000 20:15:07 -0400 $ *)
(* $I3: Copyright 1999-2002 (see COPYING for details) $ *)

val docs : (string * (string * string)) list
